from collections import namedtuple
from binascii import unhexlify

comb_config = namedtuple("comb_config",["n","t","s"])
wnaf_config = namedtuple("wnaf_config",["fixed","var"])

field_data = {
    "p25519" : {
        "gf_desc" : "2^255 - 19",
        "gf_shortname" : "25519",
        "gf_impl_bits" : 320,
        "gf_lit_limb_bits" : 51,
        "elligator_onto" : 0
    },
    "p448" : {
        "gf_desc" : "2^448 - 2^224 - 1",
        "gf_shortname" : "448",
        "gf_impl_bits" : 512,
        "gf_lit_limb_bits" : 56,
        "elligator_onto" : 0
    }
}

curve_data = {
    "curve25519" : {
        "altname" : "IsoEd25519",
        "iso_to" : "Curve25519",
        "name" : "Ristretto",
        "cofactor" : 8,
        "field" : "p25519",
        "scalar_bits" : 253,
        "d": -121665,
        "trace": -0xa6f7cef517bce6b2c09318d2e7ae9f7a,
        "mont_base": 9,
        "rist_base": "e2f2ae0a6abc4e71a884a961c500515f58e30b6aa582dd8db6a65945e08d2d76",
        "eddsa_encode_ratio": 4,
        "x_encode_ratio": 4,
        
        "combs":comb_config(3,5,17),
        "wnaf":wnaf_config(5,3),
        "window_bits":4,
        
        "eddsa_hash": "sha512",
        "eddsa_no_context": 1,
        "eddsa_dom": "SigEd25519 no Ed25519 collisions",
        "eddsa_sigma_iso": 1
    },
    "ed448goldilocks" : {
        "eddsa_encode_ratio": 4,
        "x_encode_ratio": 2,
        "altname": None,
        "name" : "Ed448-Goldilocks",
        "cofactor" : 4,
        "field" : "p448",
        "scalar_bits" : 446,
        "d": -39081,
        "trace": 0x10cd77058eec492d944a725bf7a4cf635c8e9c2ab721cf5b5529eec34,
        "rist_base": "6666666666666666666666666666666666666666666666666666666633333333333333333333333333333333333333333333333333333333",
        "mont_base": 5,
        
        "combs":comb_config(5,5,18),
        "wnaf":wnaf_config(5,3),
        "window_bits":5,
        
        "eddsa_dom":"SigEd448"
    }
}

def ser(x,bits,paren=None):
    out = ""
    mask = 2**bits - 1
    first = True
    while x > 0 or first:
        desc = "0x%0*x" % ((bits+3)//4,x&mask)
        if paren is not None:
            desc = "%s(%s)" % (paren,desc)
        if not first: out += ", "
        out += desc
        x = x >> bits
        first = False
    return out

def msqrt(x,p,hi_bit_clear = True, lo_bit_clear = False):
    if p % 4 == 3: ret = pow(x,(p+1)//4,p)
    elif p % 8 == 5:
        for u in range(1,1000):
            if pow(u,(p-1)//2,p) != 1: break
        u = pow(u,(p-1)//4,p)
        ret = pow(x,(p+3)//8,p)
        if pow(ret,2,p) != (x % p): ret = (ret * u) % p
    else: raise Exception("sqrt only for 3-mod-4 or 5-mod-8")
        
    if (ret**2-x) % p != 0: raise Exception("No sqrt")
    if hi_bit_clear and ret > p//2: ret = p-ret
    # lo_bit_clear overrides hi_bit_clear because it's not default
    if lo_bit_clear and (ret & 1): ret = p-ret 
    return ret
        
def ceil_log2(x):
    out = 0
    cmp = 1
    while x > cmp:
        cmp = cmp<<1
        out += 1
    return out

for field,data in field_data.items():
    if "modulus" not in data:
        data["modulus"] = eval(data["gf_desc"].replace("^","**"))
    
    if "gf_bits" not in data:
        data["gf_bits"] = ceil_log2(data["modulus"])

for curve,data in curve_data.items():
    for key in field_data[data["field"]]:
        if key not in data:
            data[key] = field_data[data["field"]][key]

        
    if "iso_to" not in data:
        data["iso_to"] = data["name"]
        
    if "eddsa_hash" not in data:
        data["eddsa_hash"] = "shake256"
        
    if "eddsa_no_context" not in data:
        data["eddsa_no_context"] = 0
    
    if "cxx_ns" not in data:
        data["cxx_ns"] = data["name"].replace("-","")
    
    if "eddsa_sigma_iso" not in data:
        data["eddsa_sigma_iso"] = 0
        
    if "rist_base_decoded" not in data:
        def xord(x):
            if isinstance(x,str): return ord(x)
            else: return x
        data["rist_base_decoded"] = sum(
                xord(b)<<(8*i) for i,b in enumerate(unhexlify(data["rist_base"]))
            )

    if "imagine_twist" not in data:
        # This is a HACK.  The real problem is that iso-Ed25519
        # has points at infinity unless you IMAGINE_TWIST.
        #
        # Also there are lots of bugs when cofactor=8 != IMAGINE_TWIST.
        # (FUTURE: fix all this to support other curves, eventually)
        if data["modulus"]%4 == 3: data["imagine_twist"] = 0
        else: data["imagine_twist"] = 1
        # data["imagine_twist"] = 0

    data["q"] = (data["modulus"]+1-data["trace"]) // data["cofactor"]
    data["bits"] = ceil_log2(data["modulus"])
    
    if "c_ns" not in data:
        data["c_ns"] = "decaf_" + str(data["bits"])
        data["C_NS"] = data["c_ns"].upper()

    



from textwrap import dedent
from curve_data import field_data,curve_data,ser,msqrt,ceil_log2

import os
import argparse
import re
import errno

parser = argparse.ArgumentParser(description='Generate Decaf headers and other such files.')
parser.add_argument('-o', required = True, help = "Output")
parser.add_argument('--per', required = True, help = "Files to be generated are global or per field/curve", choices=["global","field","curve"])
parser.add_argument('--item', required = False, default = "global", help = "Which curve/field to choose")
parser.add_argument('--guard', required = False, default = None, help = "header guard")
parser.add_argument('files', metavar='file', type=str, nargs='+', help='a list of files to fill')
args = parser.parse_args()

per_map = {"field":field_data, "curve":curve_data, "global":{"global":{"field":field_data,"curve":curve_data} }}

def redoc(filename,doc,author):
    doc = doc.replace("\n","\n * ")
    doc = dedent("""
        /**
         * @file %(filename)s
         * @author %(author)s
         *
         * @copyright
         *   Copyright (c) 2015-2016 Cryptography Research, Inc.  \\n
         *   Released under the MIT License.  See LICENSE.txt for license information.
         *
         * %(doc)s
         *
         * @warning This file was automatically generated in Python.
         * Please do not edit it.
         */""") % { "filename": filename, "doc": doc, "author" : author }
    doc = doc.replace(" * \n", " *\n")
    return doc[1:]

def gen_file(public,name,doc,code,per="global",author="Mike Hamburg"):
    is_header = name.endswith(".h") or name.endswith(".hxx") or name.endswith(".h++")

def fillin(template,data):
    position = 0
    ret = ""
    while True:
        dollars = template.find("$(",position)
        if dollars == -1: return ret + template[position:]
        ret += template[position:dollars]
        position = dollars + 2
        parens = 1
        while parens > 0:
            if template[position] == '(': parens += 1
            elif template[position] == ')': parens -= 1
            position += 1
        ret += str(eval(template[dollars+2:position-1],{'re':re,'ser':ser,'msqrt':msqrt,'ceil_log2':ceil_log2},data))

author = "Mike Hamburg" # FUTURE
for name in args.files:
    _,_,name_suffix = name.rpartition(".")
    template0 = open(name,"r").read()

    data = per_map[args.per][args.item]

    template = template0

    outname = args.o
    guard = args.guard
    if guard is None: guard = outname
    header_guard = "__" + guard.replace(".","_").replace("/","_").upper() + "__"

    # Extract doxygenation
    m = re.match(r"^\s*/\*\*([^*]|\*[^/])+\*/[ \t]*\n",template)
    if m:
        doc = re.sub("^\s*/?\*+/?[ \t]*","",m.group(),flags=re.MULTILINE)
        doc = re.sub("\\s*\*/","",doc)
        template = template[m.end():]
    else: doc = ""

    ns_doc = dedent(doc).strip().rstrip()
    ns_doc = redoc(guard, fillin(ns_doc,data), author)
    ns_code = fillin(template,data)
    ret = ns_doc + "\n"

    if outname.endswith(".h") or outname.endswith(".hxx"):
        ns_code = dedent("""\n
            #ifndef %(header_guard)s
            #define %(header_guard)s 1
            %(code)s
            #endif /* %(header_guard)s */
            """) % { "header_guard" : header_guard, "code": ns_code }
    ret += ns_code[1:-1]

    try:
        os.makedirs(os.path.dirname(outname))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    with open(outname,"w") as f:
        f.write(ret + "\n")



#!/usr/bin/env python
#
# Copyright 2017 John-Mark Gurney.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#

'''This is a wrapper around Ed448-Goldilocks.

This module does not follow the standard Crypto modular method
of signing due to the complexity of integration w/ the library, and
that things should be more simple to use.'''

__author__ = 'John-Mark Gurney'
__copyright__ = 'Copyright 2017 John-Mark Gurney'''
__license__ = 'BSD'
__version__ = '0.1'
__status__ = 'alpha'

import array
import os
import os.path
import sys
import unittest
import warnings

from ctypes import *

try:
	_dname = os.path.dirname(__file__)
	if not _dname:
		_dname = '.'
	_path = os.path.join(_dname, 'libdecaf.so')
	decaf = CDLL(_path)
except OSError as e: # pragma: no cover
	import warnings
	warnings.warn('libdecaf.so not installed.')
	raise ImportError(str(e))

DECAF_EDDSA_448_PUBLIC_BYTES = 57
DECAF_EDDSA_448_PRIVATE_BYTES = DECAF_EDDSA_448_PUBLIC_BYTES
DECAF_EDDSA_448_SIGNATURE_BYTES = DECAF_EDDSA_448_PUBLIC_BYTES + DECAF_EDDSA_448_PRIVATE_BYTES

# Types

ed448_pubkey_t = c_uint8 * DECAF_EDDSA_448_PUBLIC_BYTES
ed448_privkey_t = c_uint8 * DECAF_EDDSA_448_PRIVATE_BYTES
ed448_sig_t = c_uint8 * DECAF_EDDSA_448_SIGNATURE_BYTES

c_uint8_p = POINTER(c_uint8)

decaf_error_t = c_int

# Data
try:
	DECAF_ED448_NO_CONTEXT = POINTER(c_uint8).in_dll(decaf, 'DECAF_ED448_NO_CONTEXT')
except ValueError:
	DECAF_ED448_NO_CONTEXT = None

funs = {
	'decaf_ed448_derive_public_key': (None, [ ed448_pubkey_t, ed448_privkey_t]),
	'decaf_ed448_sign': (None, [ ed448_sig_t, ed448_privkey_t, ed448_pubkey_t, c_uint8_p, c_size_t, c_uint8, c_uint8_p, c_uint8 ]),
	'decaf_ed448_verify': (decaf_error_t, [ ed448_sig_t, ed448_pubkey_t, c_uint8_p, c_size_t, c_uint8, c_uint8_p, c_uint8 ]),
}

for i in funs:
	f = getattr(decaf, i)
	f.restype, f.argtypes = funs[i]

def _makeba(s):
	r = (c_ubyte * len(s))()
	r[:] = array.array('B', s)
	return r

def _makestr(a):
	# XXX - because python3 sucks, and unittest doesn't offer
	# ability to silence stupid warnings, hide the tostring
	# DeprecationWarning.
	with warnings.catch_warnings():
		warnings.simplefilter('ignore')
		return array.array('B', a).tostring()


def _ed448_privkey():
	return _makeba(os.urandom(DECAF_EDDSA_448_PRIVATE_BYTES))

class EDDSA448(object):
	_PUBLIC_SIZE = DECAF_EDDSA_448_PUBLIC_BYTES
	_PRIVATE_SIZE = DECAF_EDDSA_448_PRIVATE_BYTES
	_SIG_SIZE = DECAF_EDDSA_448_SIGNATURE_BYTES

	def __init__(self, priv=None, pub=None):
		'''Generate a new sign or verify object.  At least one
		of priv or pub MUST be specified.

		If pub is not specified, it will be generated from priv.
		If both are specified, there is no verification that pub
		is the public key for priv.

		It is recommended that you use the generate method to
		generate a new key.'''

		if priv is None and pub is None:
			raise ValueError('at least one of priv or pub must be specified.')

		if priv is not None:
			try:
				priv = _makeba(priv)
			except Exception as e:
				raise ValueError('priv must be a byte string', e)

		self._priv = priv

		if self._priv is not None and pub is None:
			self._pub = ed448_pubkey_t()
			decaf.decaf_ed448_derive_public_key(self._pub, self._priv)
		else:
			self._pub = _makeba(pub)

	@classmethod
	def generate(cls):
		'''Generate a signing object w/ a newly generated key.'''

		return cls(priv=_ed448_privkey())

	def has_private(self):
		'''Returns True if object has private key.'''

		return self._priv is not None

	def public_key(self):
		'''Returns a new object w/o the private key.  This new
		object will have the public part and can be used for
		verifying messages'''

		return self.__class__(pub=self._pub)

	def export_key(self, format):
		'''Export the key.  The only format supported is 'raw'.

		If has_private is True, then the private part will be
		exported.  If it is False, then the public part will be.
		There is no indication on the output if the key is
		public or private.  It must be tracked independantly
		of the data.'''

		if format == 'raw':
			if self._priv is None:
				return _makestr(self._pub)
			else:
				return _makestr(self._priv)
		else:
			raise ValueError('unsupported format: %s' % repr(format))

	@staticmethod
	def _makectxargs(ctx):
		if ctx is None:
			ctxargs = (DECAF_ED448_NO_CONTEXT, 0)
		else:
			ctxargs = (_makeba(ctx), len(ctx))

		return ctxargs

	def sign(self, msg, ctx=None):
		'''Returns a signature over the message.  Requires that has_private returns True.'''

		sig = ed448_sig_t()
		ctxargs = self._makectxargs(ctx)
		decaf.decaf_ed448_sign(sig, self._priv, self._pub, _makeba(msg), len(msg), 0, *ctxargs)

		return _makestr(sig)

	def verify(self, sig, msg, ctx=None):
		'''Raises an error if sig is not valid for msg.'''

		_sig = ed448_sig_t()
		_sig[:] = array.array('B', sig)
		ctxargs = self._makectxargs(ctx)
		if not decaf.decaf_ed448_verify(_sig, self._pub, _makeba(msg), len(msg), 0, *ctxargs):
			raise ValueError('signature is not valid')

def generate(curve='ed448'):
	return EDDSA448.generate()

class TestEd448(unittest.TestCase):
	def test_init(self):
		self.assertRaises(ValueError, EDDSA448)

	def test_gen(self):
		key = generate(curve='ed448')
		self.assertIsInstance(key, EDDSA448)

		self.assertTrue(key.has_private())

		pubkey = key.public_key()
		self.assertFalse(pubkey.has_private())

	def test_keyexport(self):
		# Generate key and export
		key = generate(curve='ed448')
		privkey = key.export_key('raw')

		# Generate signature
		message = b'sdlkfjsdf'
		sig = key.sign(message)

		# Verify that the key can be imported and verifies
		key2 = EDDSA448(privkey)
		key2.verify(sig, message)

		# Export the public key
		keypub = key.public_key()
		pubkey = keypub.export_key('raw')

		# Verify that the public key can be imported and verifies
		key3 = EDDSA448(pub=pubkey)
		key3.verify(sig, message)

		self.assertRaises(ValueError, key.export_key, 'PEM')

	def test_keyimportexport(self):
		privkey = b'1' * DECAF_EDDSA_448_PRIVATE_BYTES
		key = EDDSA448(privkey)

		self.assertEqual(key.export_key(format='raw'), privkey)

		key = EDDSA448(pub=b'1' * DECAF_EDDSA_448_PUBLIC_BYTES)

		self.assertRaises(ValueError, EDDSA448, priv=u'1' * DECAF_EDDSA_448_PRIVATE_BYTES)

	def test_sig(self):
		key = generate()

		message = b'this is a test message for signing'
		sig = key.sign(message)

		# Make sure sig is a string of bytes
		self.assertIsInstance(sig, bytes)
		self.assertEqual(len(sig), EDDSA448._SIG_SIZE)

		# Make sure sig is valid
		key.verify(sig, message)

		# Make sure sig is valid for public only version
		pubkey = key.public_key()
		pubkey.verify(sig, message)

		# Ensure that the wrong message fails
		message = b'this is the wrong message'
		self.assertRaises(ValueError, pubkey.verify, sig, message)

	def test_ctx(self):
		key = generate()

		message = b'foobar'
		ctx = b'contexta'
		sig = key.sign(message, ctx)

		# Make sure it verifies correctly
		key.verify(sig, message, ctx)

		# Make sure it fails w/o context
		self.assertRaises(ValueError, key.verify, sig, message)

		# Make sure it fails w/ invalid/different context
		self.assertRaises(ValueError, key.verify, sig, message, ctx + b'a')

class TestBasicLib(unittest.TestCase):
	def test_basic(self):
		priv = _ed448_privkey()
		pub = ed448_pubkey_t()

		decaf.decaf_ed448_derive_public_key(pub, priv)

		message = b'this is a test message'

		sig = ed448_sig_t()
		decaf.decaf_ed448_sign(sig, priv, pub, _makeba(message), len(message), 0, None, 0)

		r = decaf.decaf_ed448_verify(sig, pub, _makeba(message), len(message), 0, None, 0)
		self.assertTrue(r)

		message = b'aofeijseflj'
		r = decaf.decaf_ed448_verify(sig, pub, _makeba(message), len(message), 0, None, 0)
		self.assertFalse(r)



#!/usr/bin/env python

from distutils.command.build import build
from distutils.core import setup

import os

class my_build(build):
    def run(self):
        build.run(self)
        if not self.dry_run:
            os.spawnlp(os.P_WAIT, 'sh', 'sh', '-c', 'cd .. && gmake lib')
            self.copy_file(os.path.join('..', 'build', 'lib', 'libdecaf.so'), os.path.join(self.build_lib, 'edgold'))

cmdclass = {}
cmdclass['build'] = my_build

setup(name='edgold',
      version='0.1',
      description='The Ed ECC Goldilocks Python wrapper',
      author='John-Mark Gurney',
      author_email='jmg@funkthat.com',
      #url='',
      cmdclass=cmdclass,
      packages=['edgold', ],
     )
